// The -*- C++ -*- standard library header.
// This file is part of the GNU ANSI C++ Library.

#ifndef __CSTDLIB__
#define __CSTDLIB__
#include <stdlib.h>

#ifdef __GNUG__
#pragma interface "std/cstdlib.h"
#endif

extern "C++" {

#if _G_HAS_LABS
inline long   abs(long x)		{ return labs (x); }
#else
inline long   abs(long x)		{ return x >= 0 ? x : -x; }
#endif
//inline ldiv_t div(long x, long y)	{ return ldiv (x, y); }

} // extern "C++"

#endif
